/*******************************************************************************
**	Mystique generic acceleration driver
********************************************************************************
**
   Copyright (C) 1997 Jonas Borgstrm, Matthias Grimrath
**
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
** 
********************************************************************************
*/

#include <kgi/maintainers.h>
#define	MAINTAINER	Matthias_Grimrath
#define	DRIVER_NAME	"Mystique accelerator driver"
#define	DRIVER_REV	"$Revision: 1.7 $"

/*
**	KGI includes
*/

#include <kgi/module.h>
#include <kgi/kgi_commands.h>

#define MGA_CHIPSET_DRIVER /* Use our private struct definitions */

#include <Matrox/mga_common_regs.h>  /* Register definitions */
#include <Matrox/mga_common.h>
/* Driver specific things are included by mga_common_*.h */

#define ACCELIDX (PAGE_SIZE/4)
#define ACCELMSK (~PAGE_MASK/4)

const char *sugg_default[KGIGT_LAST][5] = {
	{ "vendor-matrox-mystique-t16","generic-ramdac",NULL,NULL,NULL },
	{ NULL,NULL,NULL,NULL,NULL },
	{ "generic-linear-1","generic-ramdac",NULL,NULL,NULL },
	{ "vendor-IBM-vga-4bpp",NULL,NULL,NULL,NULL},
	{ "generic-linear-8","generic-ramdac","vendor-matrox-mystique-8-15-16-32",NULL,NULL},
	{ "generic-linear-15","generic-gamma","vendor-matrox-mystique-8-15-16-32",NULL,NULL},
	{ "generic-linear-16","generic-gamma","vendor-matrox-mystique-8-15-16-32",NULL,NULL},
	{ "generic-linear-24","generic-gamma",NULL,NULL,NULL},
	{ "generic-linear-32","generic-gamma","vendor-matrox-mystique-8-15-16-32",NULL,NULL}
};

/*
** Stops the accelerator and sets DMA mode to accept FIFO commands
*/
void reset_accel(struct chipset_private *private)
{
	mem_out32(1, RESET); /* softreset engine */
	udelay(20);
	mem_out32(0, RESET);
	udelay(20);
	mem_out32(0, OPMODE); /* Reset Pseudo-DMA counter */
}

/*
** Waits for the gfx engine to finish.
*/
static int wait_accel(struct chipset_private *private)
{
	int i;
	
	/* Wait until the FIFO is empty.  Timeout value needed */
	/* because there might an unusual engine lockup */
	i = 1 + (32 - (mem_in32(FIFOSTATUS) & FIFO_COUNT_MASK)) * 300;
	while (!(mem_in32(FIFOSTATUS) & FIFO_BEMPTY) && i--) {
		udelay(20);
	}
	if (!i) {
		reset_accel(private);
		send_sig(SIGSEGV, current, 1);
	}
	i = 300;
	while ((mem_in32(STATUS) & STAT_DWGENGSTS) && i--) {
		udelay(20);
	}
	if (!i) {
		reset_accel(private);
		send_sig(SIGSEGV, current, 1);
	}
	return EOK;
}

/*
** pingpong buffer execution
*/

void pingpong_exec(struct kgi_accel *accel,struct kgi_accel_buffer *peng)
{
	struct kgi_display *dpy = accel->dpy;
	GET_PRIVATE;
	kgiu32 *acc = (kgiu32 *)peng->buffer;
	kgiu32 *a;
	kgiu32 *d;
	int count;
	int i;

	/* Polling the FIFO is necessary, otherwise the Mystique occupies
	** the WHOLE bus, and it is no longer fun to listen to your
	** soundcard while the engine draws...
	*/
	count = acc[ACCELMSK];
	if (count > 7) {
		DEBUG1("Ignoring bogus accelbuffer");
		return;
	}
	
	/* Wait until "count*4" free FIFO locations are available.
	** timeout needed, because there might be an engine lockup 
	** ??? Possible to use somehow interrupts?  According to some
	** marketing rubbish the Mystique can issue accelidle IRQs
	*/
	i = 1 + (32 - (mem_in32(FIFOSTATUS) & FIFO_COUNT_MASK)) * 300;
	while (((mem_in32(FIFOSTATUS) & FIFO_COUNT_MASK) <= count*4) && i--) {
		udelay(20);
	}
	if (!i) {
		reset_accel(private);
		send_sig(SIGSEGV, current, 1);
	}
	    
	for (a=acc,d=private->pseudo_dma.baseptr; count; count--) {
		*d++ = *a++;
		*d++ = *a++;
		*d++ = *a++;
		*d++ = *a++;
		*d++ = *a++;
	}

	peng->state = ACCEL_IDLE;

	/* !!! Once the GC is mapped, check for SYNC | ASYNC */
}

/*
** waits for the accelerator to finish
*/

void pingpong_wait(struct kgi_accel *accel)
{
	struct chipset_private *private = KGIM_PRIVATE(accel->dpy, chipset);
	wait_accel(private);
}

/*
**	KGI module interface
*/

int kgim_accel_init(struct kgi_card *card, struct kgi_display *dpy)
{
	GET_PRIVATE;

	reset_accel((struct chipset_private *)private);

	NOTICE(DRIVER_NAME " rev " DRIVER_REV " loaded");

	return EOK;
}

void kgim_accel_done(struct kgi_display *dpy)
{
	reset_accel(KGIM_PRIVATE(dpy, chipset));
	
	NOTICE("Mystique accel driver removed");
}

int kgim_accel_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
	enum kgi_tm_cmd *cmd)
{
	int chiprounding = 0;
	int i;

	switch (*cmd) {

	case CMD_CHECK:
		if (mode->request.graphtype == KGIGT_TEXT16)
			return EOK;

		if (mode->request.graphtype == KGIGT_1BIT ||
		    mode->request.graphtype == KGIGT_4BIT) {
			DEBUG1("unsupported 1,4 bpp");
			return -E(ACCEL,FAILED);
		}

		/* Visible resolution doesn't matter, virtual counts */
		/* !!! Enable lower values if accel disabled */
		switch (mode->request.graphtype) {
		case KGIGT_8BIT:
		case KGIGT_24BIT: chiprounding = 63; break;
		case KGIGT_15BIT:
		case KGIGT_16BIT:
		case KGIGT_32BIT: chiprounding = 31; break;
		default:       INTERNAL_ERROR;
		}

		if (mode->request.virt.x & chiprounding) {
			DEBUG1("virt.x %d bogus",mode->request.virt.x);
			return -E(ACCEL,FAILED);
		}

		/* Check if virt.x is suitable to enable accel */
		for (i=0; (i!=virtwidth_NUM) && 
		     (virtwidth[i] != mode->request.virt.x); i++) ;
		if (i == virtwidth_NUM) {
			DEBUG1("virt.x %d not suitable for accel",
			       mode->request.virt.x);	      
			return -E(ACCEL,FAILED);
		}

		return EOK;
		
	default:
		return EOK;
		/* return -E(ACCEL,NOSUP_ALWAYS_CANT);      */
	}
}


void kgim_accel_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	GET_PRIVATE;
	kgiu32 shift = 0;
   
	reset_accel((struct chipset_private *)private);

	switch (mode->request.graphtype) {
	case KGIGT_8BIT:  shift = 0; break;
	case KGIGT_15BIT:
	case KGIGT_16BIT: shift = 1; break;
	case KGIGT_24BIT: shift = 3; break;
	case KGIGT_32BIT: shift = 2; break;
	default:       return;
	}

	mem_out32(((mode->request.virt.x-1) << 16) | 0, CXBNDRY);
	mem_out32(0x00000000, YTOP); /* !!! respect frame # here */
	mem_out32(mode->request.virt.y * mode->request.virt.x - 1, YBOT);
	mem_out32(shift     , MACCESS);
	mem_out32(0x00000000, YDSTORG);
	mem_out32(0xFFFFFFFF, PLNWT);
	mem_out32(mode->request.virt.x, PITCH);
}

int kgim_accel_command(struct kgi_graphic *gr, kgi_uint cmd, kgi_uint arg)
{
	struct chipset_private *private;
	kgi_suggest *sugg;
	const char *name;
   
	if (!(gr->dev.state & KGI_MAPPED_ON_DISPLAY) ) {

		return -E(ACCEL,NOSUP_MODE_MMAP);
		/* Don't accel inactive screen! */
	}

	private = KGIM_PRIVATE(gr->dev.dpy, chipset);
   
	switch (cmd) { 

	/* Boring negotiation */
	case ACCEL_GETSUGGEST:
		sugg = (kgi_suggest *)(gr->io_buf);
		gr->io_res = gr->io_buf;
      
		if (sugg->id < 0 || sugg->id > 4) {

			return -E(ACCEL,BOGUS);
		}
      
		name = sugg_default[gr->dev.mode.graphtype][sugg->id];
		if (!name) {

			return -E(ACCEL,BOGUS);
		}
      
		strcpy(sugg->name, name);
		sugg->args[0] = 0;
		sugg->id++;	/* Next is RAMDAC */
		if (sugg->id >= 4 ||
		    sugg_default[gr->dev.mode.graphtype][sugg->id] == NULL) {

			sugg->id = 0;
		}
		return 0;

		/* More fun: Waiting for the accel to finish */
	case GR_MATROX_WAITACCEL:
		return wait_accel(private);

	default: 
		return -E(ACCEL,NOSUP_ALWAYS_MMAP); 
	}
}
